# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import datetime
import json
from decimal import Decimal

import django_tables2 as tables
import six
from celery.task import Task
from django.conf import settings
from django.core.cache import cache
from django.db.models import Count, Q, Sum
from django.http import (
    HttpResponse,
    HttpResponseForbidden,
    HttpResponseRedirect,
    QueryDict,
)
from django.shortcuts import get_object_or_404, render
from django.template.defaultfilters import dictsortreversed
from django.urls import reverse

from karaage.common import get_date_range
from karaage.common.decorators import admin_required, usage_required
from karaage.institutes.models import Institute
from karaage.machines.models import Account
from karaage.people.models import Person
from karaage.plugins.kgsoftware.models import Software, SoftwareVersion
from karaage.projects.models import Project

from . import graphs, models, tasks, usage
from .forms import UsageSearchForm
from .models import CPUJob
from .tables import CPUJobFilter, CPUJobTable


LOCK_EXPIRE = 60 * 60  # Lock expires in 1 hour


def progress(request):
    """ Check status of task. """
    if 'delete' in request.GET:
        models.MachineCache.objects.all().delete()
        models.InstituteCache.objects.all().delete()
        models.PersonCache.objects.all().delete()
        models.ProjectCache.objects.all().delete()
        return render(
            template_name='main.html',
            context={'content': 'Deleted'},
            request=request)

    if request.method == 'POST':
        if 'task_id' in request.POST:
            result = Task.AsyncResult(request.POST['task_id'])
            if result.failed():
                value = {
                    'info': {},
                    'ready': result.ready(),
                }
            else:
                value = {
                    'info': result.info,
                    'ready': result.ready(),
                }
            return HttpResponse(
                json.dumps(value), content_type="application/json")
    return None


def synchronise(func):
    """ If task already queued, running, or finished, don't restart. """
    def inner(request, *args):
        lock_id = '%s-%s-built-%s' % (
            datetime.date.today(), func.__name__,
            ",".join([str(a) for a in args]))

        if cache.add(lock_id, 'true', LOCK_EXPIRE):
            result = func(request, *args)
            cache.set(lock_id, result.task_id)
        else:
            task_id = cache.get(lock_id)
            if not task_id:
                return None

            cache.set(lock_id, "")
            result = Task.AsyncResult(task_id)
            if result.ready():
                result.forget()
                return None
        return result
    return inner


@synchronise
def gen_machine_category_cache(request, start, end):
    return tasks.gen_machine_category_cache.delay(start, end)


@synchronise
def gen_cache_for_machine_category(request, start, end):
    return tasks.gen_cache_for_machine_category.delay(start, end)


@synchronise
def gen_cache_for_project(request, start, end, project):
    return tasks.gen_cache_for_project.delay(start, end, project.pk)


@synchronise
def gen_cache_for_institute(request, start, end, institute):
    return tasks.gen_cache_for_institute.delay(start, end, institute.pk)


@synchronise
def gen_cache_for_all_institutes(request, start, end):
    return tasks.gen_cache_for_all_institutes.delay(start, end)


@usage_required
def index(request):
    result = progress(request)
    if result is not None:
        return result

    start, end = get_date_range(request)

    result = gen_cache_for_machine_category(request, start, end)
    if result is not None:
        return render(
            template_name='kgusage/progress.html',
            context={'task_id': result.task_id},
            request=request)

    show_zeros = True

    institute_list = Institute.active.all()
    m_list = []

    mc_cache = usage.get_machine_category_usage(start, end)
    total = mc_cache.cpu_time
    total_jobs = mc_cache.no_jobs
    available_time = mc_cache.available_time
    total_time = ((end - start).days + 1) * 24 * 60 * 60
    avg_cpus = available_time / total_time

    for m_cache in models.MachineCache.objects.filter(
            date=datetime.date.today(), start=start, end=end):
        m = m_cache.machine
        time = m_cache.cpu_time
        jobs = m_cache.no_jobs
        m_list.append({'machine': m, 'usage': time, 'jobs': jobs})

    for i_cache in models.InstituteCache.objects.filter(
            date=datetime.date.today(), start=start, end=end):
        i = i_cache.institute
        time = i_cache.cpu_time
        jobs = i_cache.no_jobs

    # Unused Entry
    unused = {'usage': available_time - total, 'quota': 0}
    if available_time != 0:
        unused['percent'] = (unused['usage'] / available_time) * 100
    else:
        unused['percent'] = 0
    unused['diff'] = unused['percent'] - unused['quota'] // 100
    if unused['diff'] <= 0:
        unused['class'] = 'green'
    else:
        unused['class'] = 'red'

    if available_time != 0:
        utilization = (Decimal(total) / available_time) * 100
    else:
        utilization = 0

    institutes_graph = graphs.get_institute_graph_url(start, end)
    machines_graph = graphs.get_machine_graph_url(start, end)
    trend_graph = graphs.get_trend_graph_url(start, end)

    return render(
        template_name='kgusage/usage_institute_list.html',
        context=locals(),
        request=request)


@usage_required
def institute_usage(request, institute_id):
    result = progress(request)
    if result is not None:
        return result

    institute = get_object_or_404(Institute, pk=institute_id)
    start, end = get_date_range(request)

    result = gen_cache_for_machine_category(request, start, end)
    if result is not None:
        return render(
            template_name='kgusage/progress.html',
            context={'task_id': result.task_id},
            request=request)

    result = gen_cache_for_institute(request, start, end, institute)
    if result is not None:
        return render(
            template_name='kgusage/progress.html',
            context={'task_id': result.task_id},
            request=request)

    project_list = []
    institute_list = Institute.active.all()

    if (not institute.can_view(request)
            and not getattr(settings, 'USAGE_IS_PUBLIC', False)):
        return HttpResponseForbidden('<h1>Access Denied</h1>')

    mc_cache = usage.get_machine_category_usage(start, end)
    available_time = mc_cache.available_time

    i_usage, i_jobs = usage.get_institute_usage(institute, start, end)

    for p_cache in models.ProjectCache.objects.filter(
            project__institute=institute,
            date=datetime.date.today(), start=start, end=end):
        p = p_cache.project
        p_usage = p_cache.cpu_time
        p_jobs = p_cache.no_jobs

    person_list = []
    person_total, person_total_jobs = 0, 0

    query = models.PersonCache.objects.filter(
        project__institute=institute,
        date=datetime.date.today(), start=start, end=end)
    query = query.order_by('-cpu_time')

    for u in query[:5]:
        person_total += u.cpu_time
        person_total_jobs += u.no_jobs
        if i_usage > 0:
            i_percent = (u.cpu_time / i_usage) * 100
        else:
            i_percent = None
        person_list.append(
            {'person': u.person,
             'project': u.project,
             'usage': u.cpu_time,
             'jobs': u.no_jobs,
             'percent': i_percent,
             })

    if i_usage > 0:
        person_percent = (person_total / i_usage) * 100
    else:
        person_percent = None

    graph = graphs.get_institute_trend_graph_url(institute, start, end)

    return render(
        template_name='kgusage/usage_institute_detail.html',
        context=locals(),
        request=request)


@usage_required
def project_usage(request, project_id):
    project = get_object_or_404(Project, pid=project_id)

    if (not project.can_view(request)
            and not getattr(settings, 'USAGE_IS_PUBLIC', False)):
        return HttpResponseForbidden('<h1>Access Denied</h1>')

    result = progress(request)
    if result is not None:
        return result

    start, end = get_date_range(request)

    result = gen_cache_for_machine_category(request, start, end)
    if result is not None:
        return render(
            template_name='kgusage/progress.html',
            context={'task_id': result.task_id},
            request=request)

    result = gen_cache_for_project(request, start, end, project)
    if result is not None:
        return render(
            template_name='kgusage/progress.html',
            context={'task_id': result.task_id},
            request=request)

    usage_list = []
    total, total_jobs = 0, 0

    # Custom SQL as need to get users that were removed from project too
    query = CPUJob.objects.filter(
        project=project,
        date__range=(start, end)
    )
    query = query.values('account').annotate().order_by('account')

    for row in query:
        u = Account.objects.get(id=row['account']).person
        time, jobs = usage.get_person_usage(u, project, start, end)
        if time:
            total += time
            total_jobs += jobs
            if jobs > 0:
                usage_list.append({'person': u, 'usage': time, 'jobs': jobs})

    for u in usage_list:
        if total == 0:
            u['percent'] = 0
        else:
            u['percent'] = (u['usage'] / total) * 100

    usage_list = dictsortreversed(usage_list, 'usage')

    count = 0
    for i in usage_list:
        i['colour'] = graphs.get_colour(count)
        count += 1

    graph = graphs.get_project_trend_graph_url(project, start, end)

    return render(
        template_name='kgusage/project_usage.html',
        context=locals(),
        request=request)


@admin_required
def unknown_usage(request):
    result = QueryDict("", mutable=True)
    result['unknown_usage'] = True
    result['sort'] = "-date"
    url = reverse('kg_usage_job_list') + "?" + result.urlencode()
    return HttpResponseRedirect(url)


@usage_required
def search(request):
    if request.method == 'POST':

        form = UsageSearchForm(request.POST)

        if form.is_valid():

            data = form.cleaned_data

            project_query = Project.objects.all()
            institute_query = Institute.objects.all()
            # person_list = Person.objects.all()

            terms = data['terms'].lower()

            start = data['start_date']
            end = data['end_date']
            machine_category = data['machine_category']
            start_str = start.strftime('%Y-%m-%d')
            end_str = end.strftime('%Y-%m-%d')
            if terms:

                # search for projects
                query = Q()
                for term in terms.split(' '):
                    q = Q(pid__icontains=term) | Q(name__icontains=term)
                    query = query & q

                project_query = project_query.filter(query)

            # search for institutes
                query = Q()
                for term in terms.split(' '):
                    q = Q(name__icontains=term)
                    query = query & q
                institute_query = institute_query.filter(query)

                project_list = []
                for p in project_query:
                    time, jobs = usage.get_project_usage(
                        p, start, end, machine_category)
                    project_list.append({
                        'obj': p,
                        'time': time,
                        'jobs': jobs,
                    })
                del project_query

                institute_list = []
                for i in institute_query:
                    time, jobs = usage.get_institute_usage(
                        i, start, end, machine_category)
                    institute_list.append({
                        'obj': i,
                        'time': time,
                        'jobs': jobs,
                    })
                del institute_query

            else:
                return HttpResponseRedirect(
                    '%s?start=%s&end=%s'
                    % (reverse('kg_usage_list'), start_str, end_str))
    else:
        start, end = get_date_range(request)
        initial = {
            'start_date': start,
            'end_date': end,
            'machine_category': request.GET.get('machine_category', None)
        }
        form = UsageSearchForm(initial=initial)

    return render(
        template_name='kgusage/search.html',
        context=locals(),
        request=request)


@usage_required
def top_users(request):
    count = 20

    result = progress(request)
    if result is not None:
        return result

    start, end = get_date_range(request)

    result = gen_cache_for_machine_category(request, start, end)
    if result is not None:
        return render(
            template_name='kgusage/progress.html',
            context={'task_id': result.task_id},
            request=request)

    start, end = get_date_range(request)
    mc_cache = usage.get_machine_category_usage(start, end)
    available_time = mc_cache.available_time
    person_list = []

    person_total, person_total_jobs = 0, 0

    query = models.PersonCache.objects.filter(
        date=datetime.date.today(), start=start, end=end,
    )

    for u in query.order_by('-cpu_time')[:count]:
        if u.cpu_time:
            person_total += u.cpu_time
            person_total_jobs += u.no_jobs
            person_list.append({
                'person': u.person,
                'project': u.project,
                'usage': u.cpu_time,
                'jobs': u.no_jobs,
                'percent': ((u.cpu_time / available_time) * 100)
            })

    person_percent = (person_total / available_time) * 100

    return render(
        template_name='kgusage/top_users.html',
        context=locals(),
        request=request)


@usage_required
def institute_trends(request):
    result = progress(request)
    if result is not None:
        return result

    start, end = get_date_range(request)

    result = gen_cache_for_all_institutes(request, start, end)
    if result is not None:
        return render(
            template_name='kgusage/progress.html',
            context={'task_id': result.task_id},
            request=request)

    graph_list = graphs.get_institutes_trend_graph_urls(start, end)

    return render(
        template_name='kgusage/institute_trends.html',
        context=locals(),
        request=request)


@usage_required
def institute_users(request, institute_id):
    result = progress(request)
    if result is not None:
        return result

    start, end = get_date_range(request)

    result = gen_cache_for_machine_category(request, start, end)
    if result is not None:
        return render(
            template_name='kgusage/progress.html',
            context={'task_id': result.task_id},
            request=request)

    institute = get_object_or_404(Institute, pk=institute_id)

    if (not institute.can_view(request)
            and not getattr(settings, 'USAGE_IS_PUBLIC', False)):
        return HttpResponseForbidden('<h1>Access Denied</h1>')

    start, end = get_date_range(request)

    mc_cache = usage.get_machine_category_usage(start, end)
    available_time = mc_cache.available_time

    person_list = []

    query = models.PersonCache.objects.filter(
        date=datetime.date.today(), start=start, end=end,
        person__institute=institute,
        no_jobs__gt=0)

    person_total, person_total_jobs = 0, 0
    for u in query.order_by('-cpu_time'):
        person_total = person_total + u.cpu_time
        person_total_jobs = person_total_jobs + u.no_jobs
        person_list.append({
            'person': u.person,
            'project': u.project,
            'usage': u.cpu_time,
            'jobs': u.no_jobs,
            'percent': ((u.cpu_time / available_time) * 100),
        })

    if available_time != 0:
        person_percent = (person_total / available_time) * 100
    else:
        person_percent = 0

    return render(
        template_name='kgusage/institute_users.html',
        context=locals(),
        request=request)


@usage_required
def core_report(request):
    start, end = get_date_range(request)

    job_list = CPUJob.objects.filter(date__gte=start, date__lte=end)

    core_1 = job_list.filter(cores=1).count()
    core_2_4 = job_list.filter(cores__gte=2, cores__lte=4).count()
    core_5_8 = job_list.filter(cores__gte=5, cores__lte=8).count()
    core_9_16 = job_list.filter(cores__gte=9, cores__lte=16).count()
    core_17_32 = job_list.filter(cores__gte=17, cores__lte=32).count()
    core_33_64 = job_list.filter(cores__gte=33, cores__lte=64).count()
    core_65_128 = job_list.filter(cores__gte=65, cores__lte=128).count()
    core_128 = job_list.filter(cores__gte=128).count()

    data = [core_1, core_2_4, core_5_8, core_9_16,
            core_17_32, core_33_64, core_65_128, core_128]
    total = sum(data)

#    x_labels = ['1', '2-4', '5-8', '9-16', '17-32', '33-64', '65-128', '128+']
#    max_y = max(data)
#    data = {'Total jobs': data}
#    g = GraphGenerator()
#    graph = g.bar_chart(data, x_labels, max_y, bar_width=50).get_url()

    return render(
        template_name='kgusage/core_report.html',
        context=locals(),
        request=request)


@usage_required
def mem_report(request):

    start, end = get_date_range(request)

    job_list = CPUJob.objects.filter(date__gte=start, date__lte=end)

    mem_0_4 = job_list.filter(mem__lte=4 * 1024 * 1024).count()
    mem_4_8 = job_list.filter(
        mem__gt=4 * 1024 * 1024, mem__lte=8 * 1024 * 1024).count()
    mem_8_16 = job_list.filter(
        mem__gt=8 * 1024 * 1024, mem__lte=16 * 1024 * 1024).count()
    mem_16_32 = job_list.filter(
        mem__gt=16 * 1024 * 1024, mem__lte=32 * 1024 * 1024).count()
    mem_32_64 = job_list.filter(
        mem__gt=32 * 1024 * 1024, mem__lte=64 * 1024 * 1024).count()
    mem_64_128 = job_list.filter(
        mem__gt=64 * 1024 * 1024, mem__lte=128 * 1024 * 1024).count()
    mem_128 = job_list.filter(mem__gt=128 * 1024 * 1024).count()
    data = [mem_0_4, mem_4_8, mem_8_16, mem_16_32,
            mem_32_64, mem_64_128, mem_128]
    total = sum(data)

#    x_labels = ['0-4', '4-8', '8-16', '16-32', '32-64', '64-128', '128+']
#    labels = []
#    max_y = max(data)
#    data = {'Total jobs': data}
#    g = GraphGenerator()
#    graph = g.bar_chart(data, x_labels, max_y, bar_width=50).get_url()

    return render(
        template_name='kgusage/mem_report.html',
        context=locals(),
        request=request)


@usage_required
def job_detail(request, jobid):
    job = get_object_or_404(CPUJob, jobid=jobid)

    if (not job.project.can_view(request)
            and not getattr(settings, 'USAGE_IS_PUBLIC', False)):
        return HttpResponseForbidden('<h1>Access Denied</h1>')

    return render(
        template_name='kgusage/job_detail.html',
        context={'job': job},
        request=request)


@usage_required
def job_list(request):
    if not getattr(settings, 'USAGE_IS_PUBLIC', False):
        return HttpResponseForbidden('<h1>Access Denied</h1>')

    queryset = CPUJob.objects.select_related()

    q_filter = CPUJobFilter(request.GET, queryset=queryset)
    table = CPUJobTable(q_filter.qs)
    tables.RequestConfig(request).configure(table)

    spec = []
    for name, value in six.iteritems(q_filter.form.cleaned_data):
        if value is not None and value != "":
            name = name.replace('_', ' ').capitalize()
            spec.append((name, value))

    return render(
        template_name='kgusage/job_list.html',
        context={
            'table': table,
            'filter': q_filter,
            'spec': spec,
            'title': "Job list",
        },
        request=request)


@admin_required
def software_stats(request, software_id):
    software = get_object_or_404(Software, pk=software_id)
    start, end = get_date_range(request)
    querystring = request.META.get('QUERY_STRING', '')
    if software.softwareversion_set.count() == 1:
        sv = software.softwareversion_set.all()[0]
        url = reverse('kg_software_version_stats', args=[sv.id])
        return HttpResponseRedirect(url)
    version_stats = SoftwareVersion.objects \
        .filter(software=software, cpujob__date__range=(start, end)) \
        .annotate(jobs=Count('cpujob'), usage=Sum('cpujob__cpu_usage')) \
        .filter(usage__isnull=False)
    version_totaljobs = version_stats.aggregate(Sum('jobs'))['jobs__sum']
    # version_totalusage = version_stats.aggregate(Sum('usage'))
    person_stats = Person.objects \
        .filter(account__cpujob__software__software=software,
                account__cpujob__date__range=(start, end)) \
        .annotate(jobs=Count('account__cpujob'),
                  p_usage=Sum('account__cpujob__cpu_usage'))
    project_stats = Project.objects \
        .filter(cpujob__software__software=software,
                cpujob__date__range=(start, end)) \
        .annotate(jobs=Count('cpujob'),
                  p_usage=Sum('cpujob__cpu_usage'))
    person_institute_stats = Institute.objects \
        .filter(person__account__cpujob__software__software=software,
                person__account__cpujob__date__range=(start, end)) \
        .annotate(jobs=Count('person__account__cpujob'),
                  i_usage=Sum('person__account__cpujob__cpu_usage'))
    project_institute_stats = Institute.objects \
        .filter(project__cpujob__software__software=software,
                project__cpujob__date__range=(start, end)) \
        .annotate(jobs=Count('project__cpujob'),
                  i_usage=Sum('project__cpujob__cpu_usage'))

    context = {
        'software': software,
        'version_stats': version_stats,
        'version_totaljobs': version_totaljobs,
        'person_stats': person_stats,
        'person_institute_stats': person_institute_stats,
        'project_stats': project_stats,
        'project_institute_stats': project_institute_stats,
        'start': start,
        'end': end,
        'querystring': querystring,
    }
    return render(
        template_name='kgusage/software_stats.html',
        context=context,
        request=request)


@admin_required
def version_stats(request, version_id):
    version = get_object_or_404(SoftwareVersion, pk=version_id)
    start, end = get_date_range(request)
    querystring = request.META.get('QUERY_STRING', '')

    person_stats = Person.objects \
        .filter(account__cpujob__software=version,
                account__cpujob__date__range=(start, end)) \
        .annotate(jobs=Count('account__cpujob'),
                  p_usage=Sum('account__cpujob__cpu_usage'))
    project_stats = Project.objects \
        .filter(cpujob__software=version,
                cpujob__date__range=(start, end)) \
        .annotate(jobs=Count('cpujob'),
                  p_usage=Sum('cpujob__cpu_usage'))
    person_institute_stats = Institute.objects \
        .filter(person__account__cpujob__software=version,
                person__account__cpujob__date__range=(start, end)) \
        .annotate(jobs=Count('person__account__cpujob'),
                  i_usage=Sum('person__account__cpujob__cpu_usage'))
    project_institute_stats = Institute.objects \
        .filter(project__cpujob__software=version,
                project__cpujob__date__range=(start, end)) \
        .annotate(jobs=Count('project__cpujob'),
                  i_usage=Sum('project__cpujob__cpu_usage'))

    context = {
        'version': version,
        'person_stats': person_stats,
        'person_institute_stats': person_institute_stats,
        'project_stats': project_stats,
        'project_institute_stats': project_institute_stats,
        'start': start,
        'end': end,
        'querystring': querystring,
    }

    return render(
        template_name='kgusage/version_stats.html',
        context=context,
        request=request)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

# Django 1.6 Hack: Ensure tldap.django gets initialised.
import django
from django.utils.deprecation import MiddlewareMixin


if django.VERSION < (1, 7):
    import tldap.django  # NOQA

# threadlocals middleware
try:
    from threading import local
except ImportError:
    from django.utils._threading_local import local


_thread_locals = local()


def get_current_user():
    return getattr(_thread_locals, 'user', None)


def reset():
    _thread_locals.user = None


class ThreadLocals(MiddlewareMixin):
    """Middleware that gets various objects from the
    request object and saves them in thread local storage."""

    def process_request(self, request):
        _thread_locals.user = getattr(request, 'user', None)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from django.core.management import call_command
from django.test import TestCase
from mock import Mock

from karaage.signals import daily_cleanup


class CommandsTestCase(TestCase):

    def test_daily_cleanup(self):
        callback = Mock()
        daily_cleanup.connect(callback)

        try:
            call_command('daily_cleanup')
        finally:
            daily_cleanup.disconnect(daily_cleanup)

        self.assertEqual(callback.call_count, 1)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import django_filters
import django_tables2 as tables
from django_tables2.utils import A

from .models import Software, SoftwareLicenseAgreement


class SoftwareFilter(django_filters.FilterSet):
    description = django_filters.CharFilter(lookup_expr="icontains")
    begin__last_used = django_filters.DateFilter(
        field_name="softwareversion__last_used",
        lookup_expr="gte")
    end_last_used = django_filters.DateFilter(
        field_name="softwareversion__last_used",
        lookup_expr="lte")

    class Meta:
        model = Software
        fields = ('name', 'description', 'group', 'category', 'academic_only',
                  'restricted',)


class SoftwareTable(tables.Table):
    name = tables.LinkColumn('kg_software_detail', args=[A('pk')])
    group = tables.LinkColumn('kg_group_detail', args=[A('group.name')])
    softwareversion__last_used = tables.Column(verbose_name="Last used")

    class Meta:
        model = Software
        fields = ('name', 'description', 'group', 'category',
                  'softwareversion__last_used')
        empty_text = "No items"


class SoftwareLicenseAgreementTable(tables.Table):
    software = tables.LinkColumn(
        'kg_software_detail', accessor="license.software",
        args=[A('license.software.pk')])
    license = tables.LinkColumn(
        'kg_software_license_detail', args=[A('license.pk')])
    person = tables.LinkColumn(
        'kg_person_detail', args=[A('person.username')])

    class Meta:
        model = SoftwareLicenseAgreement
        fields = ("software", "license", "person", "date")
        empty_text = "No items"



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import datetime

import pytest
from django.core.management import call_command
from django.test import TestCase
from django.urls import reverse

from karaage.machines.models import Machine
from karaage.middleware.threadlocals import reset
from karaage.people.models import Person
from karaage.projects.models import Project


@pytest.mark.django_db
class AccountTestCase(TestCase):

    def setUp(self):
        def cleanup():
            reset()
        self.addCleanup(cleanup)

        call_command('loaddata', 'test_karaage', **{'verbosity': 0})
        form_data = {
            'title': 'Mr',
            'short_name': 'Sam',
            'full_name': 'Sam Morrison2',
            'position': 'Sys Admin',
            'institute': 1,
            'department': 'eddf',
            'email': 'sam2@vpac.org',
            'country': 'AU',
            'telephone': '4444444',
            'username': 'samtest2',
            'password1': 'Exaiquouxei0',
            'password2': 'Exaiquouxei0',
            'needs_account': False,
        }
        self.client.login(username='kgsuper', password='aq12ws')
        response = self.client.post(reverse('kg_person_add'), form_data)
        self.assertEqual(response.status_code, 302)

        person = Person.objects.get(username='kgsuper')
        self.assertEqual(person.short_name, 'Super')
        self.assertEqual(person.full_name, 'Super User')

    def test_add_account(self):
        project = Project.objects.get(pk=1)
        person = Person.objects.get(username="samtest2")
        person.groups.add(project.group)

        response = self.client.get(
            reverse('kg_account_add', args=['samtest2']))
        self.assertEqual(response.status_code, 200)

        form_data = {
            'username': person.username,
            'shell': '/bin/bash',
            'default_project': 1,
        }

        response = self.client.post(
            reverse('kg_account_add', args=['samtest2']), form_data)
        self.assertEqual(response.status_code, 302)
        person = Person.objects.get(username="samtest2")
        self.assertTrue(person.has_account())

    def test_fail_add_accounts_username(self):
        project = Project.objects.get(pk=1)
        person = Person.objects.get(username="samtest2")
        person.groups.add(project.group)

        form_data = {
            'username': person.username,
            'shell': '/bin/bash',
            'default_project': 1,
        }
        response = self.client.post(
            reverse('kg_account_add', args=['samtest2']), form_data)
        self.assertEqual(response.status_code, 302)

        response = self.client.post(
            reverse('kg_account_add', args=['samtest2']), form_data)
        self.assertContains(
            response, "Username already in use.")

    def test_fail_add_accounts_project(self):
        form_data = {
            'username': 'samtest2',
            'shell': '/bin/bash',
            'default_project': 1,
        }
        response = self.client.post(
            reverse('kg_account_add', args=['samtest2']), form_data)
        self.assertContains(
            response, "Person does not belong to default project")

        project = Project.objects.get(pk=1)
        person = Person.objects.get(username="samtest2")
        person.groups.add(project.group)

        form_data = {
            'username': person.username,
            'shell': '/bin/bash',
            'default_project': 1,
        }
        response = self.client.post(
            reverse('kg_account_add', args=['samtest2']), form_data)
        self.assertEqual(response.status_code, 302)

        form_data = {
            'username': person.username,
            'shell': '/bin/bash',
            'default_project': 1,
        }

        response = self.client.post(
            reverse('kg_account_add', args=['samtest2']), form_data)
        self.assertContains(
            response, "Username already in use.")

    def test_lock_unlock_account(self):
        project = Project.objects.get(pk=1)
        person = Person.objects.get(username="samtest2")
        person.groups.add(project.group)

        response = self.client.get(reverse('kg_account_add',
                                           args=['samtest2']))
        self.assertEqual(response.status_code, 200)

        form_data = {
            'username': person.username,
            'shell': '/bin/bash',
            'default_project': 1,
        }

        response = self.client.post(
            reverse('kg_account_add', args=['samtest2']), form_data)
        self.assertEqual(response.status_code, 302)
        person = Person.objects.get(username='samtest2')
        ua = person.get_account()
        self.assertEqual(person.is_locked(), False)
        self.assertEqual(ua.login_shell(), '/bin/bash')

        response = self.client.post(
            reverse('kg_person_lock', args=['samtest2']))
        person = Person.objects.get(username='samtest2')
        ua = person.get_account()
        self.assertEqual(person.is_locked(), True)
        self.assertEqual(ua.login_shell(), '/bin/bash')

        response = self.client.post(
            reverse('kg_person_unlock', args=['samtest2']))
        person = Person.objects.get(username='samtest2')
        ua = person.get_account()
        self.assertEqual(person.is_locked(), False)
        self.assertEqual(ua.login_shell(), '/bin/bash')


@pytest.mark.django_db
class MachineTestCase(TestCase):

    def setUp(self):
        today = datetime.datetime.now()
        # 10cpus
        mach1 = Machine.objects.get(pk=1)
        mach1.start_date = today - datetime.timedelta(days=80)
        mach1.save()
        # 40 cpus
        mach2 = Machine.objects.get(pk=2)
        mach2.start_date = today - datetime.timedelta(days=100)
        mach2.end_date = today - datetime.timedelta(days=20)
        mach2.save()
        # 8000 cpus
        mach3 = Machine.objects.get(pk=3)
        mach3.start_date = today - datetime.timedelta(days=30)
        mach3.save()



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from django.conf import settings
from django.contrib.auth.tokens import default_token_generator
from django.template.loader import render_to_string
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode

from karaage.common import log
from karaage.common.emails import CONTEXT, send_mail


def render_email(name, context):
    subject = render_to_string(
        ['karaage/emails/%s_subject.txt' % name,
            'karaage/people/emails/%s_subject.txt' % name],
        context)
    subject = subject.replace('\n', '')
    body = render_to_string(
        ['karaage/emails/%s_body.txt' % name,
            'karaage/people/emails/%s_body.txt' % name],
        context)
    return subject, body


def send_bounced_warning(person, leader_list):
    """Sends an email to each project leader for person
    informing them that person's email has bounced"""
    context = CONTEXT.copy()
    context['person'] = person

    for lp in leader_list:
        leader = lp['leader']

        context['project'] = lp['project']
        context['receiver'] = leader

        to_email = leader.email
        subject = render_to_string(
            'karaage/people/emails/bounced_email_subject.txt', context)
        body = render_to_string(
            'karaage/people/emails/bounced_email_body.txt', context)
        send_mail(
            subject.replace('\n', ''), body,
            settings.ACCOUNTS_EMAIL, [to_email])
        log.change(
            leader,
            'Sent email about bounced emails from %s' % person)


def send_reset_password_email(person):
    """Sends an email to user allowing them to set their password."""
    uid = urlsafe_base64_encode(force_bytes(person.pk)).decode("ascii")
    token = default_token_generator.make_token(person)
    url = '%s/persons/reset/%s/%s/' % (
        settings.REGISTRATION_BASE_URL, uid, token)

    context = CONTEXT.copy()
    context.update({
        'url': url,
        'receiver': person,
    })

    to_email = person.email
    subject, body = render_email('reset_password', context)

    send_mail(subject, body, settings.ACCOUNTS_EMAIL, [to_email])


def send_confirm_password_email(person):
    """Sends an email to user allowing them to confirm their password."""
    url = '%s/profile/login/%s/' % (
        settings.REGISTRATION_BASE_URL, person.username)

    context = CONTEXT.copy()
    context.update({
        'url': url,
        'receiver': person,
    })

    to_email = person.email
    subject, body = render_email('confirm_password', context)

    send_mail(subject, body, settings.ACCOUNTS_EMAIL, [to_email])



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

try:
    from factory.django import DjangoModelFactory
    from factory.fuzzy import FuzzyText
    import factory
except ImportError:
    raise ImportError(
        "factory_boy is required, "
        "either install from a package or using \'pip install -e .[tests]\'")

from karaage.tests.fixtures import (
    InstituteFactory,
    PersonFactory,
    ProjectFactory,
    fuzzy_lower_text,
)

from ..models import Application, ProjectApplication


class ApplicationFactory(DjangoModelFactory):
    applicant = factory.SubFactory(PersonFactory)

    class Meta:
        model = Application


class ProjectApplicationFactory(ApplicationFactory):

    class Meta:
        model = ProjectApplication


class NewProjectApplicationFactory(ProjectApplicationFactory):
    name = fuzzy_lower_text(prefix='projectapplication-')
    description = FuzzyText()
    institute = factory.SubFactory(InstituteFactory)


class ExistingProjectApplicationFactory(ProjectApplicationFactory):
    project = factory.SubFactory(ProjectFactory)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


def fix_applications(apps, schema_editor):
    content_type = apps.get_model("contenttypes", "ContentType")
    content_type.objects.filter(model="softwareapplication").update(app_label="kgsoftware.applications")


class Migration(migrations.Migration):

    dependencies = [
        ('kgsoftware', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(fix_applications),
    ]



import datetime

import pytest
from django.test import TestCase

from karaage.machines.models import Machine


@pytest.mark.django_db
class MachineTestCase(TestCase):
    fixtures = [
        'test_karaage.json',
        'test_kgusage.json',
    ]

    def setUp(self):
        today = datetime.datetime.now()
        # 10cpus
        mach1 = Machine.objects.get(pk=1)
        mach1.start_date = today - datetime.timedelta(days=80)
        mach1.save()
        # 40 cpus
        mach2 = Machine.objects.get(pk=2)
        mach2.start_date = today - datetime.timedelta(days=100)
        mach2.end_date = today - datetime.timedelta(days=20)
        mach2.save()
        # 8000 cpus
        mach3 = Machine.objects.get(pk=3)
        mach3.start_date = today - datetime.timedelta(days=30)
        mach3.save()

    def do_availablity_test(self, start, end, expected_time, expected_cpu):
        from karaage.plugins.kgusage.usage import get_machine_category_usage
        cache = get_machine_category_usage(start.date(), end.date())
        available_time = cache.available_time
        self.assertEqual(available_time, expected_time)

    def fixme_test_available_time(self):
        day = 60 * 60 * 24
        today = datetime.datetime.now()

        end = today - datetime.timedelta(days=20)
        start = today - datetime.timedelta(days=30)
        self.do_availablity_test(start, end, 8050 * day * 11, 8050)

        start = today - datetime.timedelta(days=99)
        end = today - datetime.timedelta(days=90)
        self.do_availablity_test(start, end, 40 * day * 10, 40)

        start = today - datetime.timedelta(days=85)
        end = today - datetime.timedelta(days=76)
        self.do_availablity_test(start, end, 45 * day * 10, 45)

        start = today - datetime.timedelta(days=35)
        end = today - datetime.timedelta(days=16)
        self.do_availablity_test(start, end, 6042 * day * 20, 6042)

        start = today - datetime.timedelta(days=20)
        end = today - datetime.timedelta(days=20)
        self.do_availablity_test(start, end, 8050 * day, 8050)



"""
Sphinx plugins for Karaage documentation. Copied from Django.
"""
import json
import os
import re

from docutils import nodes
from docutils.parsers.rst import Directive
from sphinx import addnodes
from sphinx.builders.html import StandaloneHTMLBuilder
from sphinx.domains.std import Cmdoption
from sphinx.util import logging
from sphinx.util.console import bold
from sphinx.writers.html import HTMLTranslator

logger = logging.getLogger(__name__)
# RE for option descriptions without a '--' prefix
simple_option_desc_re = re.compile(
    r'([-_a-zA-Z0-9]+)(\s*.*?)(?=,\s+(?:/|-|--)|$)')


def setup(app):
    app.add_crossref_type(
        directivename="setting",
        rolename="setting",
        indextemplate="pair: %s; setting",
    )
    app.add_crossref_type(
        directivename="templatetag",
        rolename="ttag",
        indextemplate="pair: %s; template tag"
    )
    app.add_crossref_type(
        directivename="templatefilter",
        rolename="tfilter",
        indextemplate="pair: %s; template filter"
    )
    app.add_crossref_type(
        directivename="fieldlookup",
        rolename="lookup",
        indextemplate="pair: %s; field lookup type",
    )
    app.add_object_type(
        directivename="django-admin",
        rolename="djadmin",
        indextemplate="pair: %s; django-admin command",
        parse_node=parse_django_admin_node,
    )
    app.add_directive('django-admin-option', Cmdoption)
    app.add_config_value('karaage_next_version', '0.0', True)
    app.add_directive('versionadded', VersionDirective)
    app.add_directive('versionchanged', VersionDirective)
    app.add_builder(KaraageStandaloneHTMLBuilder)


class VersionDirective(Directive):
    has_content = True
    required_arguments = 1
    optional_arguments = 1
    final_argument_whitespace = True
    option_spec = {}

    def run(self):
        if len(self.arguments) > 1:
            msg = """Only one argument accepted for directive '{directive_name}::'.
            Comments should be provided as content,
            not as an extra argument.""".format(directive_name=self.name)
            raise self.error(msg)

        env = self.state.document.settings.env
        ret = []
        node = addnodes.versionmodified()
        ret.append(node)

        if self.arguments[0] == env.config.karaage_next_version:
            node['version'] = "Development version"
        else:
            node['version'] = self.arguments[0]

        node['type'] = self.name
        if self.content:
            self.state.nested_parse(self.content, self.content_offset, node)
        try:
            env.get_domain('changeset').note_changeset(node)
        except ExtensionError:
            # Sphinx < 1.8: Domain 'changeset' is not registered
            env.note_versionchange(node['type'], node['version'], node, self.lineno)
        return ret


class KaraageHTMLTranslator(HTMLTranslator):
    """
    Karaage-specific reST to HTML tweaks.
    """

    # Don't use border=1, which docutils does by default.
    def visit_table(self, node):
        self.context.append(self.compact_p)
        self.compact_p = True
        self._table_row_index = 0  # Needed by Sphinx
        self.body.append(self.starttag(node, 'table', CLASS='docutils'))

    def depart_table(self, node):
        self.compact_p = self.context.pop()
        self.body.append('</table>\n')

    def visit_desc_parameterlist(self, node):
        self.body.append('(')  # by default sphinx puts <big> around the "("
        self.first_param = 1
        self.optional_param_level = 0
        self.param_separator = node.child_text_separator
        self.required_params_left = sum(isinstance(c, addnodes.desc_parameter) for c in node.children)

    def depart_desc_parameterlist(self, node):
        self.body.append(')')

    #
    # Turn the "new in version" stuff (versionadded/versionchanged) into a
    # better callout -- the Sphinx default is just a little span,
    # which is a bit less obvious that I'd like.
    #
    # FIXME: these messages are all hardcoded in English. We need to change
    # that to accommodate other language docs, but I can't work out how to make
    # that work.
    #
    version_text = {
        'versionchanged': 'Changed in Karaage %s',
        'versionadded': 'New in Karaage %s',
    }

    def visit_versionmodified(self, node):
        self.body.append(
            self.starttag(node, 'div', CLASS=node['type'])
        )
        version_text = self.version_text.get(node['type'])
        if version_text:
            title = "%s%s" % (
                version_text % node['version'],
                ":" if node else "."
            )
            self.body.append('<span class="title">%s</span> ' % title)

    def depart_versionmodified(self, node):
        self.body.append("</div>\n")

    # Give each section a unique ID -- nice for custom CSS hooks
    def visit_section(self, node):
        old_ids = node.get('ids', [])
        node['ids'] = ['s-' + i for i in old_ids]
        node['ids'].extend(old_ids)
        super().visit_section(node)
        node['ids'] = old_ids


def parse_django_admin_node(env, sig, signode):
    command = sig.split(' ')[0]
    env.ref_context['std:program'] = command
    title = "kg-manage %s" % sig
    signode += addnodes.desc_name(title, title)
    return command


class KaraageStandaloneHTMLBuilder(StandaloneHTMLBuilder):
    """
    Subclass to add some extra things we need.
    """

    name = 'djangohtml'

    def finish(self):
        super().finish()
        logger.info(bold("writing templatebuiltins.js..."))
        xrefs = self.env.domaindata["std"]["objects"]
        templatebuiltins = {
            "ttags": [
                n for ((t, n), (k, a)) in xrefs.items()
                if t == "templatetag" and k == "ref/templates/builtins"
            ],
            "tfilters": [
                n for ((t, n), (k, a)) in xrefs.items()
                if t == "templatefilter" and k == "ref/templates/builtins"
            ],
        }
        outfilename = os.path.join(self.outdir, "templatebuiltins.js")
        with open(outfilename, 'w') as fp:
            fp.write('var django_template_builtins = ')
            json.dump(templatebuiltins, fp)
            fp.write(';\n')



from celery import shared_task
from django.core import management

from karaage.signals import daily_cleanup


@shared_task
def daily():
    management.call_command('lock_expired')
    daily_cleanup.send(sender=daily)



#! /usr/bin/env python
######################################################################
#  Copyright (c) by Kevin L. Sitze on 2011-01-23.
#  This code may be used pursuant to the MIT License.
######################################################################

# Source:
# http://code.activestate.com/recipes/577551-trace-decorator-for-debugging/

"""This package provides facilities to attach decorators to classes or
modules (possibly recursively).  A tracing decorator is provided for
tracing function and method calls in your applications.

from decorators import *

@trace
def function(...):
    '''Output goes to <TT>logging.get_logger(__module__)</TT>,
    where <TT>__module__</TT> is the name of the module in which
    the function is defined.  If the current module is '__main__'
    then the root logger will be used.
    '''
    pass

@trace(logger)
def function(...):
    '''Output goes to the specified logger instance.
    '''
    pass

@trace("Logger")
def function(...):
    '''Output goes to the named logger instance.
    '''

class ClassName(object):
    '''Output goes to the ClassName logger.  If ClassName is
    not in the '__main__' module then the name of its module
    will be prefixed to the classname (separated by a dot).
    All methods on the class will be traced.
    '''
    __logger__ = loggerOrLoggerName # optional
    __metaclass__ = decorators.TraceMetaClass
    def method(...):
        pass

class PostDecorate(object):
    def method(...):
        pass
# attach a decorator to an existing class.
attach(trace(logger), PostDecorate)

You can also attach a decorator to an existing module.
"""
from __future__ import absolute_import, print_function

import inspect
import logging
import sys
import types
from functools import wraps
from itertools import chain

import six


try:
    import _thread as thread
except ImportError:
    import thread


FunctionTypes = tuple({types.BuiltinFunctionType, types.FunctionType})

MethodTypes = tuple({types.BuiltinMethodType, types.MethodType})


class _C(object):

    @classmethod
    def class_method(cls):
        pass
    classMethodType = type(class_method)

    @staticmethod
    def static_method():
        pass
    staticMethodType = type(static_method)

    @property
    def property_method(self):
        return None


ClassMethodType = _C.classMethodType
StaticMethodType = _C.staticMethodType
PropertyType = type(_C.property_method)

CallableTypes = tuple({
    types.BuiltinFunctionType, types.FunctionType,
    types.BuiltinMethodType, types.MethodType, ClassMethodType
})

__all__ = (
    'ThreadLocal',
    'TraceMetaClass',
    'attach',
    'get_formatter',
    'set_formatter',
    'get_logger_factory',
    'set_logger_factory',
    'trace'
)


######################################################################
#  Utility functions
######################################################################

MAX_SIZE = 320


def chop(value):
    s = repr(value)
    if len(s) > MAX_SIZE:
        return s[:MAX_SIZE] + '...' + s[-1]
    else:
        return s


def loggable(obj):
    """Return "True" if the obj implements the minimum Logger API
    required by the 'trace' decorator.
    """
    if isinstance(obj, logging.Logger):
        return True
    else:
        return (inspect.isclass(obj)
                and inspect.ismethod(getattr(obj, 'debug', None))
                and inspect.ismethod(getattr(obj, 'isEnabledFor', None))
                and inspect.ismethod(getattr(obj, 'setLevel', None)))


######################################################################
#  LoggerFactory Property
######################################################################

_logger_factory = logging


def get_logger_factory():
    """Retrieve the current factory object for creating loggers.
    The default is to use the logging module.
    """
    global _logger_factory
    return _logger_factory


def set_logger_factory(factory):
    """Set a factory object for creating loggers.  This object must
    publish a method or class named 'get_logger' that takes a string
    parameter naming the logger instance to retrieve.  Logger objects
    returned by this factory must, at a minimum, expose the methods
    'isEnabledFor' and 'debug'.
    """
    global _logger_factory
    _logger_factory = factory


######################################################################
#  class PrependLoggerFactory
######################################################################

class PrependLoggerFactory(object):

    """This is a convenience class for creating new loggers for the
    "trace" decorator.  All loggers created via this class will have a
    user specified prefix prepended to the name of the logger to
    instantiate.
    """

    def __init__(self, prefix='trace'):
        """Construct a new "PrependLoggerFactory" instance that
        prepends the value \var{prefix} to the name of each
        logger to be created by this class.
        """
        self.__prefix = prefix.strip('.')

    @property
    def prefix(self):
        """The value to prefix to each logger created by this factory.
        """
        return self.__prefix

    @prefix.setter
    def prefix(self, value):
        self.__prefix = value.strip('.')

    def get_logger(self, name):
        return logging.getLogger('.'.join((self.__prefix, name)))


######################################################################
#  class ThreadLocal
######################################################################

class ThreadLocal(object):

    """Instances of this class provide a thread-local variable.
    """

    def __init__(self):
        self.__lock = thread.allocate_lock()
        self.__vars = dict()
        self.__init = None

    @property
    def value(self):
        with self.__lock:
            try:
                return self.__vars[thread.get_ident()]
            except KeyError:
                return self.__init

    @value.setter
    def value(self, value):
        with self.__lock:
            self.__vars[thread.get_ident()] = value

    @property
    def initial_value(self):
        with self.__lock:
            return self.__init

    @initial_value.setter
    def initial_value(self, value):
        with self.__lock:
            self.__init = value


######################################################################
#  Formatter functions
######################################################################

def _formatter_self(name, value):
    """Format the "self" variable and value on instance methods.
    """
    __mname = value.__module__
    if __mname != '__main__':
        return '%s = <%s.%s object at 0x%x>' \
            % (name, __mname, value.__class__.__name__, id(value))
    else:
        return '%s = <%s object at 0x%x>' \
            % (name, value.__class__.__name__, id(value))


def _formatter_class(name, value):
    """Format the "klass" variable and value on class methods.
    """
    __mname = value.__module__
    if __mname != '__main__':
        return "%s = <type '%s.%s'>" % (name, __mname, value.__name__)
    else:
        return "%s = <type '%s'>" % (name, value.__name__)


def _formatter_named(name, value):
    """Format a named parameter and its value.
    """
    return '%s = %s' % (name, chop(value))


def _formatter_defaults(name, value):
    return '[%s = %s]' % (name, chop(value))


af_self = _formatter_self
af_class = _formatter_class
af_named = _formatter_named
af_default = _formatter_defaults
af_unnamed = chop
af_keyword = _formatter_named


def get_formatter(name):
    """Return the named formatter function.  See the function
    "set_formatter" for details.
    """
    if name in ('self', 'instance', 'this'):
        return af_self
    elif name == 'class':
        return af_class
    elif name in ('named', 'param', 'parameter'):
        return af_named
    elif name in ('default', 'optional'):
        return af_default
#    elif name in ('anonymous', 'arbitrary', 'unnamed'):
#        return af_anonymous
    elif name in ('keyword', 'pair', 'pairs'):
        return af_keyword
    else:
        raise ValueError('unknown trace formatter %r' % name)


def set_formatter(name, func):
    """Replace the formatter function used by the trace decorator to
    handle formatting a specific kind of argument.  There are several
    kinds of arguments that trace discriminates between:

    * instance argument - the object bound to an instance method.
    * class argument - the class object bound to a class method.
    * positional arguments (named) - values bound to distinct names.
    * positional arguments (default) - named positional arguments with
      default values specified in the function declaration.
    * positional arguments (anonymous) - an arbitrary number of values
      that are all bound to the '*' variable.
    * keyword arguments - zero or more name-value pairs that are
      placed in a dictionary and bound to the double-star variable.

    \var{name} - specifies the name of the formatter to be modified.

        * instance argument - "self", "instance" or "this"
        * class argument - "class"
        * named argument - "named", "param" or "parameter"
        * default argument - "default", "optional"
        * anonymous argument - "anonymous", "arbitrary" or "unnamed"
        * keyword argument - "keyword", "pair" or "pairs"

    \var{func} - a function to format an argument.
    * For all but anonymous formatters this function must accept two
      arguments: the variable name and the value to which it is bound.
    * The anonymous formatter function is passed only one argument
      corresponding to an anonymous value.
    * if \var{func} is "None" then the default formatter will be used.
    """
    if name in ('self', 'instance', 'this'):
        global af_self
        af_self = _formatter_self if func is None else func
    elif name == 'class':
        global af_class
        af_class = _formatter_class if func is None else func
    elif name in ('named', 'param', 'parameter'):
        global af_named
        af_named = _formatter_named if func is None else func
    elif name in ('default', 'optional'):
        global af_default
        af_default = _formatter_defaults if func is None else func
    elif name in ('anonymous', 'arbitrary', 'unnamed'):
        global af_anonymous
        af_anonymous = chop if func is None else func
    elif name in ('keyword', 'pair', 'pairs'):
        global af_keyword
        af_keyword = _formatter_named if func is None else func
    else:
        raise ValueError('unknown trace formatter %r' % name)


######################################################################
#  Decorator: trace
######################################################################

__builtins = (
    '__import__(name,globals={},locals={},fromlist=[],level=-1)',
    'abs(number)',
    'all(iterable)',
    'any(iterable)',
    'apply(object,args=[],kwargs={})',
    'bin(number)',
    'callable(object)',
    'chr(i)',
    'cmp(x,y)',
    'coerce(x,y)',
    'compile(source,filename,mode,flags=0,dont_inherit=0)',
    'delattr(object,name)',
    'dir()',
    'divmod(x,y)',
    'eval(source,globals={},locals={})',
    'execfile(filename,globals={},locals={})',
    'filter(function,sequence)',
    'format(value,format_spec="")',
    'getattr(object,name)',
    'globals()',
    'hasattr(object,name)',
    'hash(object)',
    'hex(number)',
    'id(object)',
    'input(prompt=None)',
    'intern(string)',
    'isinstance(object,klass)',
    'issubclass(C,B)',
    'iter(collection)',
    'len(object)',
    'locals()',
    'map(function,sequence)',
    'max(iterable,key=None)',
    'min(iterable,key=None)',
    'next(iterator)',
    'oct(number)',
    'open(name,mode=0666,buffering=True)',
    'ord(c)',
    'pow(x,y,z=None)',
    'range()',
    'raw_input(prompt=None)',
    'reduce(function,sequence)',
    'reload(module)',
    'repr(object)',
    'round(number,ndigits=0)',
    'setattr(object,name,value)',
    'sorted(iterable,cmp=None,key=None,reverse=False)',
    'sum(sequence,start=0)',
    'unichr(i)',
    'vars()',
    'zip(sequence)'
)

__builtin_defaults = {
    '""': "",
    '-1': -1,
    '0': 0,
    '0666': 0o666,
    'False': False,
    'None': None,
    'True': True,
    '[]': list(),
    '{}': dict()
}

__builtin_functions = None


def __lookup_builtin(name):
    """Lookup the parameter name and default parameter values for
    builtin functions.
    """
    global __builtin_functions
    if __builtin_functions is None:
        builtins = dict()
        for proto in __builtins:
            pos = proto.find('(')
            name, params, defaults = proto[:pos], list(), dict()
            for param in proto[pos + 1:-1].split(','):
                pos = param.find('=')
                if not pos < 0:
                    param, value = param[:pos], param[pos + 1:]
                    try:
                        defaults[param] = __builtin_defaults[value]
                    except KeyError:
                        raise ValueError(
                            'builtin function %s: parameter %s: '
                            'unknown default %r' % (name, param, value))
                params.append(param)
            builtins[name] = (params, defaults)
        __builtin_functions = builtins

    try:
        params, defaults = __builtin_functions[name]
    except KeyError:
        params, defaults = tuple(), dict()
        __builtin_functions[name] = (params, defaults)
        print(
            "Warning: builtin function %r is missing prototype" % name,
            file=sys.stderr)
    return len(params), params, defaults


_ = ThreadLocal()
_.initial_value = False


def trace(_name):
    """Function decorator that logs function entry and exit details.

    \var{_name} a string, an instance of logging.Logger or a function.

    Construct a function or method proxy to generate call traces.
    """
    def decorator(_func):
        """This is the actual decorator function that wraps the
        \var{_func} function for detailed logging.
        """

        def positional(name, value):
            """Format one named positional argument.
            """
            if name is __self:
                return af_self(name, value)
            elif name is __klass:
                return af_class(name, value)
            else:
                return af_named(name, value)

        def wrapper(*__argv, **__kwds):
            if not logger.isEnabledFor(logging.DEBUG) or _.value:
                return _func(*__argv, **__kwds)

            try:
                _.value = True
                params = dict(co_defaults)
                params.update(__kwds)
                params.update(zip(co_varnames, __argv))
                if 'raw_password' in params:
                    params['raw_password'] = '<censored>'

                position = [
                    positional(n, params.pop(n))
                    for n in co_varnames[:len(__argv)]
                ]
                defaults = [
                    af_default(n, params.pop(n))
                    for n in co_varnames[len(__argv):]
                ]
                nameless = (
                    af_unnamed(v) for v in __argv[co_argcount:]
                )
                keywords = (
                    af_keyword(n, params[n]) for n in sorted(params.keys())
                )

                params = ', '.join(
                    filter(None, chain(
                        position,
                        defaults,
                        nameless,
                        keywords)))
                # params = params.replace(', [', '[, ').replace('][, ', ', ')

                enter = [pre_enter]
                if params:
                    enter.append(' ')
                    enter.append(params)
                    enter.append(' ')
                enter.append(')')

                leave = [pre_leave]

                try:
                    logger.debug(''.join(enter))
                    try:
                        try:
                            _.value = False
                            result = _func(*__argv, **__kwds)
                        finally:
                            _.value = True
                    except Exception:
                        ex_type, value, traceback = sys.exc_info()
                        leave.append(' => exception thrown\n\traise ')
                        __mname = ex_type.__module__
                        if __mname != '__main__':
                            leave.append(__mname)
                            leave.append('.')
                        leave.append(ex_type.__name__)
                        if value.args:
                            leave.append('(')
                            leave.append(
                                ', '.join(chop(v) for v in value.args))
                            leave.append(')')
                        else:
                            leave.append('()')
                        raise
                    else:
                        if result is not None:
                            leave.append(' => ')
                            leave.append(chop(result))
                finally:
                    logger.debug(''.join(leave))
            finally:
                _.value = False

            return result

        ####
        #  decorator
        ####

        __self = False
        __klass = False

        def nop(x):
            return x

        __rewrap = nop
        if type(_func) in FunctionTypes:
            # functions do not belong to a class.
            __cname = None
        elif type(_func) in MethodTypes:
            # im_self is None for unbound instance methods.
            # Assumption: trace is only called on unbound methods.
            if _func.im_self is not None:
                __rewrap = classmethod
                __cname = _func.im_self.__name__
                __klass = True
            else:
                __cname = _func.im_class.__name__
                __self = True
            _func = _func.im_func
        else:
            # other callables are not supported yet.
            return _func
        __module = _func.__module__
        __fname = _func.__name__

        # Do not wrap initialization and conversion methods.
        if __fname in ('__init__', '__new__', '__repr__', '__str__'):
            return __rewrap(_func)

        # Generate the Fully Qualified Function Name.

        __fqfn = list()
        if __module != '__main__':
            __fqfn.append(__module)
            __fqfn.append('.')
        if __cname is not None:
            __fqfn.append(__cname)
            __fqfn.append('.')
        __fqfn.append(__fname)
        __fqfn = ''.join(__fqfn)

        if type(_name) in CallableTypes:
            logger = get_logger_factory().get_logger(__fqfn)
        elif loggable(_name):
            logger = _name
        elif isinstance(_name, six.string_types):
            logger = get_logger_factory().get_logger(_name)
        else:
            raise ValueError(
                'invalid object %r: must be a function, a method, '
                'a string or an object that implements the Logger API' % _name)

        pre_enter = ['>>> ', __fqfn, '(']
        pre_enter = ''.join(pre_enter)

        pre_leave = ['<<< ', __fqfn]
        pre_leave = ''.join(pre_leave)

        ####
        #  Here we are really mucking around in function internals.
        #  func_code is the low level 'code' instance that describes
        #  the function arguments, variable and other stuff.
        #
        #  func.func_code.co_argcount - number of function arguments.
        #  func.func_code.co_varnames - function variables names, the
        #      first co_argcount values are the argument names.
        #  func.func_defaults - contains default arguments

        try:
            code = _func.__code__
        except AttributeError:
            co_argcount, co_varnames, co_defaults = \
                __lookup_builtin(_func.__name__)
        else:
            co_argcount = code.co_argcount
            co_varnames = code.co_varnames[:co_argcount]
            if _func.__defaults__:
                co_defaults = dict(
                    zip(co_varnames[-len(_func.__defaults__):],
                        _func.__defaults__))
            else:
                co_defaults = dict()
            if __klass:
                __klass = co_varnames[0]
            if __self:
                __self = co_varnames[0]
        return __rewrap(wraps(_func)(wrapper))

    ####
    #  trace
    ####

#    logging.basicConfig(level = logging.DEBUG)
    if type(_name) in CallableTypes:
        return decorator(_name)
    else:
        return decorator


######################################################################
#  attach: apply decorator to a class or module
######################################################################

def attach_to_property(decorator, klass, k, prop_attr, prop_decorator):
    if prop_attr is not None:
        setattr(klass, k, prop_attr)
        value = decorator(getattr(klass, k))
    else:
        value = None
    # Passing "None" to the property decorator causes the new property
    # to assume the original value of the associated attribute.
    return prop_decorator(value)


def attach_to_class(decorator, klass, recursive=True):
    for k, v in six.iteritems(klass.__dict__):
        t = type(v)
        if isinstance(v, types.FunctionType) or t is ClassMethodType:
            setattr(klass, k, decorator(getattr(klass, k)))
        elif t is StaticMethodType:
            setattr(klass, k, staticmethod(decorator(getattr(klass, k))))
        elif t is PropertyType:
            value = getattr(klass, k)
            value = attach_to_property(
                decorator, klass, k, value.fget, value.getter)
            value = attach_to_property(
                decorator, klass, k, value.fset, value.setter)
            value = attach_to_property(
                decorator, klass, k, value.fdel, value.deleter)
            setattr(klass, k, value)
        elif recursive and inspect.isclass(v):
            attach_to_class(decorator, v, recursive)


def attach(decorator, obj, recursive=True):
    """attach(decorator, class_or_module[, recursive = True])

    Utility to attach a \val{decorator} to the \val{obj} instance.

    If \val{obj} is a module, the decorator will be attached to every
    function and class in the module.

    If \val{obj} is a class, the decorator will be attached to every
    method and subclass of the class.

    if \val{recursive} is "True" then subclasses will be decorated.
    """
    if inspect.ismodule(obj):
        for name, fn in inspect.getmembers(obj, inspect.isfunction):
            setattr(obj, name, decorator(fn))
        for name, klass in inspect.getmembers(obj, inspect.isclass):
            attach_to_class(decorator, klass, recursive)
    elif inspect.isclass(obj):
        attach_to_class(decorator, obj, recursive)


######################################################################
#  class TraceMetaClass
######################################################################

class TraceMetaClass(type):

    """Metaclass to automatically attach the 'trace' decorator to all
    methods, static method and class methods of the class.
    """
    def __new__(mcs, class_name, bases, class_dict):
        klass = super(TraceMetaClass, mcs) \
            .__new__(mcs, class_name, bases, class_dict)
        if '__logger__' in class_dict:
            hook = trace(class_dict['__logger__'])
        else:
            hook = trace
        attach_to_class(hook, klass, False)
        return klass


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    logger = logging.root

    class Test(object):
        __logger__ = logger
        __metaclass__ = TraceMetaClass

        @classmethod
        def class_method(cls):
            pass

        @staticmethod
        def static_method():
            pass

        __test = None

        @property
        def test(self):
            return self.__test

        @test.setter
        def test(self, value):
            self.__test = value

        def method(self):
            pass

    Test.class_method()
    Test.static_method()
    test = Test()
    test.test = 1
    assert 1 == test.test
    test.method()

    class Test(object):

        @classmethod
        def class_method(cls):
            pass

        @staticmethod
        def static_method():
            pass

        __test = None

        @property
        def test(self):
            return self.__test

        @test.setter
        def test(self, value):
            self.__test = value

        def method(self):
            pass

        def __str__(self):
            return 'Test(' + str(self.test) + ')'

    attach(trace(logger), Test)
    Test.class_method()
    Test.static_method()
    test = Test()
    test.test = 1
    assert 1 == test.test
    test.method()
    print(str(test))

    @trace(logger)
    def test(x, y, z=True):
        a = x + y
        b = x * y
        if z:
            return a
        else:
            return b

    test(5, 5)
    test(5, 5, False)

    set_logger_factory(PrependLoggerFactory())

    @trace('main')
    def test(x, *argv, **kwds):
        """Simple test
        """
        return x + sum(argv)

    test(5)
    test(5, 5)
    test(5, 5, False)
    test(5, 5, False, name=10)

    test(*range(50))

    assert test.__doc__ == 'Simple test\n        '
    assert test.__name__ == 'test'

    myzip = trace('main')(zip)
    for i, j in myzip(range(5), range(5, 10)):
        print(i, j)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from django.conf import settings
from django.conf.urls import url

from karaage.machines.views import accounts


urlpatterns = [
    url(r'^add/(?P<username>%s)/$' % settings.USERNAME_VALIDATION_RE,
        accounts.add_account, name='kg_account_add'),
    url(r'^add_project/(?P<username>%s)/$' % settings.USERNAME_VALIDATION_RE,
        accounts.add_project, name='kg_account_add_project'),
    url(r'^(?P<account_id>\d+)/$', accounts.account_detail,
        name='kg_account_detail'),
    url(r'^(?P<account_id>\d+)/edit/$', accounts.edit_account,
        name='kg_account_edit'),
    url(r'^(?P<account_id>\d+)/delete/$', accounts.delete_account,
        name='kg_account_delete'),
    url(r'^(?P<account_id>\d+)/makedefault/(?P<project_id>%s)/$'
        % settings.PROJECT_VALIDATION_RE,
        accounts.make_default, name='kg_account_set_default'),

    url(r'^no_project/$', accounts.no_project_list,
        name='kg_account_no_project'),
    url(r'^no_default/$', accounts.no_default_list,
        name='kg_account_no_default'),
    url(r'^wrong_default/$', accounts.wrong_default_list,
        name='kg_account_wrong_default'),
    url(r'^no_account/$', accounts.no_account_list,
        name='kg_account_no_account'),
]



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.
from django.contrib.humanize.templatetags.humanize import intcomma
from django.template import Library
from django.template.defaultfilters import filesizeformat


register = Library()


@register.filter
def timeformat(value):

    if value == '':
        return ''

    if value is None:
        return '0s'
    if value < 60:
        return '%ss' % intcomma(int(value))
    # less than 1 hour
    elif value < 3600:
        v = int(value / 60)
        return '%sm' % intcomma(v)
    # less than 1 day
    # elif value < 86400:
    #    v = int(value/3600)
    #    return '%sh' % intcomma(v)
    # less than a month
    # elif value < 2592000:
    #    v = int(value/86400)
    #    return '%sd' % intcomma(v)
    # less than 1 year
    # elif value < 31104000:
    #    v = int(value/2592000)
    #    return '%smonth' % intcomma(v)
    else:
        v = int(value / 3600)
        return '%sh' % intcomma(v)


@register.filter
def fileformat(kilobytes):
    if kilobytes:
        return filesizeformat(kilobytes * 1024)
    return None



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import django_tables2 as tables
import six
from django.conf import settings
from django.db.models import Q
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse

from karaage.common import is_admin
from karaage.common.decorators import admin_required
from karaage.common.models import LogEntry
from karaage.common.tables import LogEntryFilter, LogEntryTable
from karaage.people.models import Group, Person
from karaage.people.tables import GroupTable, PersonTable
from karaage.projects.models import Project
from karaage.projects.tables import ProjectTable


@admin_required
def admin_index(request):
    config = tables.RequestConfig(request, paginate={"per_page": 5})

    newest_people = Person.objects.order_by('-date_approved', '-id')
    newest_people = newest_people.filter(date_approved__isnull=False)
    newest_people = newest_people.select_related()
    newest_people = PersonTable(
        newest_people, orderable=False, prefix="people-")
    config.configure(newest_people)

    newest_projects = Project.objects.order_by('-date_approved')
    newest_projects = newest_projects.filter(date_approved__isnull=False)
    newest_projects = newest_projects.filter(is_active=True)
    newest_projects = newest_projects.select_related()
    newest_projects = ProjectTable(
        newest_projects, orderable=False, prefix="projects-")
    config.configure(newest_projects)

    recent_actions = LogEntry.objects.order_by('-action_time', '-id')
    recent_actions = LogEntryTable(
        recent_actions, orderable=False, prefix="actions-")
    config.configure(recent_actions)

    var = {
        'newest_people': newest_people,
        'newest_projects': newest_projects,
        'recent_actions': recent_actions,
    }
    return render(
        template_name='karaage/common/index.html', context=var,
        request=request)


def index(request):
    if settings.ADMIN_REQUIRED or is_admin(request):
        return admin_index(request)
    return render(
        template_name='karaage/common/index.html', request=request)


@admin_required
def search(request):
    config = tables.RequestConfig(request, paginate={"per_page": 5})

    if 'sitesearch' in request.GET and request.GET['sitesearch'].strip() != "":
        people_list = Person.objects.all()
        group_list = Group.objects.all()
        project_list = Project.objects.all()

        new_data = request.GET.copy()
        siteterms = new_data['sitesearch'].lower()
        term_list = siteterms.split(' ')

        # people
        query = Q()
        for term in term_list:
            q = Q(username__icontains=term)
            q = q | Q(short_name__icontains=term)
            q = q | Q(full_name__icontains=term)
            q = q | Q(email__icontains=term)
            query = query & q

        people_list = people_list.filter(query).distinct()
        people_list = PersonTable(people_list, prefix="people-")
        config.configure(people_list)

        # groups
        query = Q()
        for term in term_list:
            q = Q(name__icontains=term) | Q(description__icontains=term)
            query = query & q

        group_list = group_list.filter(query)
        group_list = GroupTable(group_list, prefix="group-")
        config.configure(group_list)

        # projects
        query = Q()
        for term in term_list:
            q = Q(pid__icontains=term)
            q = q | Q(name__icontains=term)
            q = q | Q(leaders__username__icontains=term)
            q = q | Q(leaders__short_name__icontains=term)
            q = q | Q(leaders__full_name__icontains=term)
            query = query & q

        project_list = project_list.filter(query).distinct()
        project_list = ProjectTable(project_list, prefix="project-")
        config.configure(project_list)

        return render(
            template_name='karaage/common/site_search.html',
            context=locals(),
            request=request)
    else:
        return HttpResponseRedirect(reverse('index'))


@admin_required
def log_list(request):
    queryset = LogEntry.objects.all()

    q_filter = LogEntryFilter(request.GET, queryset=queryset)
    table = LogEntryTable(q_filter.qs)
    tables.RequestConfig(request).configure(table)

    spec = []
    for name, value in six.iteritems(q_filter.form.cleaned_data):
        if value is not None and value != "":
            name = name.replace('_', ' ').capitalize()
            spec.append((name, value))

    return render(
        template_name='karaage/common/log_list.html',
        context={
            'table': table,
            'filter': q_filter,
            'spec': spec,
            'title': "Institute list",
        },
        request=request)


@admin_required
def misc(request):
    from karaage.common.simple import direct_to_template
    return direct_to_template(
        request,
        template='karaage/common/misc_detail.html')


def aup(request):
    from karaage.common.simple import direct_to_template
    return direct_to_template(
        request,
        template='karaage/common/aup_detail.html')



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

"""
Methods for getting usage data
uses a database cache to speed up proccess

"""
import datetime

from .models import (
    InstituteCache,
    MachineCache,
    MachineCategoryCache,
    PersonCache,
    ProjectCache,
)


def get_institute_usage(institute, start, end):
    """Return a tuple of cpu hours and number of jobs for an institute
    for a given period

    Keyword arguments:
    institute --
    start -- start date
    end -- end date
    """
    try:
        cache = InstituteCache.objects.get(
            institute=institute, date=datetime.date.today(),
            start=start, end=end)
        return cache.cpu_time, cache.no_jobs
    except InstituteCache.DoesNotExist:
        return 0, 0


def get_project_usage(project, start, end):
    """Return a tuple of cpu hours and number of jobs for a project
    for a given period

    Keyword arguments:
    project --
    start -- start date
    end -- end date

    """
    try:
        cache = ProjectCache.objects.get(
            project=project, date=datetime.date.today(),
            start=start, end=end)
        return cache.cpu_time, cache.no_jobs
    except ProjectCache.DoesNotExist:
        return 0, 0


def get_person_usage(person, project, start, end):
    """Return a tuple of cpu hours and number of jobs for a person in a
    specific project

    Keyword arguments:
    person --
    project -- The project the usage is from
    start -- start date
    end -- end date
    """
    try:
        cache = PersonCache.objects.get(
            person=person, project=project, date=datetime.date.today(),
            start=start, end=end)
        return cache.cpu_time, cache.no_jobs
    except PersonCache.DoesNotExist:
        return 0, 0


def get_machine_usage(machine, start, end):
    """Return a tuple of cpu hours and number of jobs for a machine
    for a given period

    Keyword arguments:
    machine --
    start -- start date
    end -- end date

    """

    try:
        cache = MachineCache.objects.get(
            machine=machine, date=datetime.date.today(),
            start=start, end=end)
        return cache.cpu_time, cache.no_jobs
    except MachineCache.DoesNotExist:
        return 0, 0


def get_machine_category_usage(start, end):
    """Return a tuple of cpu hours and number of jobs
    for a given period

    Keyword arguments:
    start -- start date
    end -- end date

    """

    cache = MachineCategoryCache.objects.get(
        date=datetime.date.today(),
        start=start, end=end)
    return cache



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import ajax_select
import six
from django.core.exceptions import PermissionDenied
from django.db.models import Q
from django.utils.html import escape

from karaage.common import is_admin
from karaage.projects.models import Project


class LookupChannel(ajax_select.LookupChannel):
    """ Base clase for lookups. """

    def check_auth(self, request):
        """
        to ensure that nobody can get your data via json simply by knowing the
        URL.  public facing forms should write a custom LookupChannel to
        implement as you wish.  also you could choose to return
        HttpResponseForbidden("who are you?") instead of raising
        PermissionDenied (401 response)
        """
        if not request.user.is_authenticated:
            raise PermissionDenied
        if not is_admin(request):
            raise PermissionDenied


class ProjectLookup(LookupChannel):
    model = Project

    def get_query(self, q, request):
        """ return a query set searching for the query string q
            either implement this method yourself or set the search_field
            in the LookupChannel class definition
        """
        return Project.objects.filter(
            Q(pid__icontains=q)
            | Q(name__icontains=q)
        )

    def get_result(self, obj):
        """
        result is the simple text that is the completion of what the person
        typed
        """
        return obj.pid

    def format_match(self, obj):
        """
        (HTML) formatted item for display in the dropdown
        """
        return escape(six.u("%s") % obj)

    def format_item_display(self, obj):
        """
        (HTML) formatted item for displaying item in the selected deck area
        """
        return escape(six.u("%s") % obj)
